/*
* Copyright (C) 1999 Winston Chang
*                    <winstonc@cs.wisc.edu>
*                    <winston@stdout.org>
* Copyright (C) 2009 elsamuko
*                    <elsamuko@web.de>
* Copyright (C) 2009 Søren Hauberg
*                    <hauberg@gmail.com>
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
  Simple GIMP-octave terminal interface, compile and install with:

    make
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <sys/param.h>

#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

#include <vte/vte.h>

#define PLUG_IN_PROC    "gimp-octave-terminal"
#define PLUG_IN_BINARY  "gimp-octave-terminal"

#define WORKING_DIR ".gimp-octave-terminal"
#define MATRIX_IN   "matrix_in.txt"
#define MATRIX_OUT  "matrix_out.txt"

typedef struct
{
  char wd [PATH_MAX + 1];
  char cinput [PATH_MAX + 1];
  char coutput [PATH_MAX + 1];
} OctaveParams;


typedef struct
{
  gboolean  run;
} OctaveInterface;


/* local function prototypes */
inline
gint coord (gint x, gint y, gint k, gint channels, gint width)
{
  return channels * (width * y + x) + k;
}

void reset_default (void);

static
void query (void);

static
void run (const gchar *name, gint nparams, const GimpParam *param,
          gint *nreturn_vals, GimpParam **return_vals);
 
static
void save_data (char *filename, GimpDrawable *drawable);
                
static
gboolean write_octave_image (char *filename, guchar *layer, gint width, gint height,
                             gint bpp);

static
gboolean read_octave_image (char *filename, gpointer preview);

static
gboolean octave_dialog (GimpDrawable *drawable);

static
void preview_update (GimpPreview *preview);

/* create a few globals, set default values */
static OctaveParams octave_params =
{
  "", /* default input  */
  "", /* default output */
};

guchar *result_layer = NULL;

/* Setting PLUG_IN_INFO */
const GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

MAIN ();

static
void query (void)
{
  static const
  GimpParamDef args [] =
    {
      { GIMP_PDB_INT32,    (gchar*)"run-mode",  (gchar*)"The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }" },
      { GIMP_PDB_IMAGE,    (gchar*)"image",     (gchar*)"Image" },
      { GIMP_PDB_DRAWABLE, (gchar*)"drawable",  (gchar*)"Drawable to draw on" },
      { GIMP_PDB_STRING,   (gchar*)"input",     (gchar*)"Filename of the input matrix (nyi)" },
      { GIMP_PDB_STRING,   (gchar*)"output",    (gchar*)"Filename of the output matrix (nyi)" },
      { GIMP_PDB_STRING,   (gchar*)"script",    (gchar*)"Filename of the Octave script (nyi)" }
    };

  gimp_install_procedure (PLUG_IN_PROC,
                          "Edit images with Gnu Octave from GIMP",
                          "Edit images with Gnu Octave from GIMP",
                          "elsamuko <elsamuko@web.de>", /* XXX: What should be here? */
                          "elsamuko",
                          "1999-2009",
                          "_Octave Terminal",
                          "GRAY*, RGB*", /* XXX: do we actually handle alpha channels */
                          GIMP_PLUGIN,
                          G_N_ELEMENTS (args), 0,
                          args, NULL);

  gimp_plugin_menu_register (PLUG_IN_PROC, "<Image>/Filters/Mathematics");
}


static
void run (const gchar *name, gint nparams, const GimpParam *param, gint *nreturn_vals,
          GimpParam **return_vals)
{
  reset_default ();

  static GimpParam values [1];
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;
  GimpDrawable *drawable;
  GimpRunMode run_mode;

  run_mode = (GimpRunMode)param[0].data.d_int32;

  *return_vals = values;
  *nreturn_vals = 1;

  values[0].type = GIMP_PDB_STATUS;
  values[0].data.d_status = status;

  /* Get drawable information... */
  drawable = gimp_drawable_get (param[2].data.d_drawable);
  gimp_tile_cache_ntiles (2 * MAX (drawable->width  / gimp_tile_width () + 1,
                                   drawable->height / gimp_tile_height () + 1));

  /* Write image to disc */
  gimp_get_data (PLUG_IN_PROC, &octave_params);
  drawable = gimp_drawable_get (param[2].data.d_drawable);
  save_data (octave_params.cinput, drawable);

  switch (run_mode)
    {
      case GIMP_RUN_INTERACTIVE:
        if (! octave_dialog (drawable))
          return;
        break;

      default:
        break;
    }

  if (status == GIMP_PDB_SUCCESS)
    {
      drawable = gimp_drawable_get (param[2].data.d_drawable);

      gimp_displays_flush ();

      gimp_drawable_detach (drawable);
      values[0].data.d_status = status;
    }
}

static
void save_data (char *filename, GimpDrawable *drawable)
{
  GimpPixelRgn PR;
  gint width, height, bpp;

  width = drawable->width;
  height = drawable->height;
  bpp = drawable->bpp;

  /* Initialize pixel regions */
  gimp_pixel_rgn_init (&PR, drawable, 0, 0, width, height, FALSE, FALSE);

  /* Save */
  guchar *layer;
  layer = g_new (guchar, bpp * width * height);
  gimp_pixel_rgn_get_rect (&PR, layer, 0, 0, width, height);
  write_octave_image (filename, layer, width, height, bpp);
  g_free (layer);
}

static
gboolean write_octave_image (char *filename, guchar *layer, gint width, gint height,
                             gint bpp)
{
  FILE *fid;
  gboolean error;
  gint k, i, j;
  
  error = FALSE;
  fid = fopen (filename, "w");

  if (fid != NULL)
    {
      /* Write header */
      fputs ("# Created by gimp-Octave-terminal\n", fid);
      fputs ("# name: im\n", fid);
      fputs ("# type: matrix\n", fid);
      if (bpp == 1)
        {
          fprintf (fid, "# ndims: 2\n");
          fprintf (fid, "%i %i\n", height, width);
        }
      else
        {
          fprintf (fid, "# ndims: 3\n");
          fprintf (fid, "%i %i %i\n", height, width, bpp);
        }

      for (k = 0; k < bpp; k++)
        for (i = 0; i < width; i++)
          for (j = 0; j < height; j++)
            fprintf (fid, "%i\n", (gint)layer [coord (i, j, k, bpp, width)]);
      fclose (fid);
    }
  else
    {
      gimp_message ("Error: Could not open input matrix.");
      printf ("L%i: Error: Could not open input matrix.", __LINE__);
      error = TRUE;
    }
  
  return error;
}

static
gboolean read_octave_image (char *filename, gpointer preview)
{
  FILE *fid;
  gint bpp, width, height;
  gint i, j, k, num;
  char str [101];
  char first;
  //guchar *layer;
  gboolean success;
  
  success = TRUE;
  fid = fopen (filename, "r");
  
  if (fid != NULL)
    {
      /* Skip header */
      first = '#';
      while (first == '#')
        {
          fgets (str, 100, fid);
          first = str [0];
        }
      
      /* Read size info */
      sscanf (str, "%i %i %i", &height, &width, &bpp);
      
      if (result_layer != NULL)
        g_free (result_layer);
      result_layer = g_new (guchar, bpp * width * height);

      /* Read data */
      if (result_layer != NULL)
        {
          for (k = 0; k < bpp; k++)
            for (i = 0; i < width; i++)
              for (j = 0; j < height; j++)
                {
                  fscanf (fid, "%i", &num);
                  result_layer [coord (i, j, k, bpp, width)] = num;
                }

          GimpDrawable *drawable;
          GimpPixelRgn PR;

          drawable = gimp_drawable_preview_get_drawable (GIMP_DRAWABLE_PREVIEW (preview));

          gimp_pixel_rgn_init (&PR, drawable, 0, 0, width, height, TRUE, TRUE);

          gimp_pixel_rgn_set_rect (&PR, result_layer, 0, 0, width, height);
          gimp_drawable_preview_draw_region (GIMP_DRAWABLE_PREVIEW (preview), &PR);

          gimp_drawable_flush (drawable);
          gimp_drawable_merge_shadow (drawable->drawable_id, TRUE);
          gimp_drawable_update (drawable->drawable_id, 0, 0, width, height);
        }
      else
        success = FALSE;

      fclose (fid);
    }
  else
    {
      gimp_message ("Error: Could not open return matrix.");
      printf ("L%i: Error: Could not open return matrix.", __LINE__);
      success = FALSE;
    }
  
  return success;
}

static
void on_octave_exit (VteTerminal *vteterminal, gpointer dialog)
{
  gtk_widget_destroy (GTK_WIDGET (GIMP_DIALOG (dialog))); // XXX: This generates an error
}

static
void on_file_change (GFileMonitor *monitor, GFile *file, GFile *other_file,
                     GFileMonitorEvent event_type, gpointer preview)
{
  if (event_type == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT)
    {
      read_octave_image (octave_params.coutput, preview);
      //gimp_preview_invalidate (GIMP_PREVIEW (preview));
    }
}

static
gboolean octave_dialog (GimpDrawable *drawable)
{
  GtkWidget *dialog;
  GtkWidget *main_hbox;
  GtkWidget *preview;
  GtkWidget *terminal;
  gboolean run;
  pid_t pid;

  /* default entries: */
  reset_default ();

  gimp_ui_init (PLUG_IN_BINARY, TRUE);

  dialog = gimp_dialog_new ("GNU Octave Terminal", PLUG_IN_BINARY, NULL, (GtkDialogFlags)0,
                            gimp_standard_help_func, PLUG_IN_PROC, GTK_STOCK_CANCEL,
                            GTK_RESPONSE_CANCEL, GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);

  gimp_window_set_transient (GTK_WINDOW (dialog));

  main_hbox = gtk_hbox_new (FALSE, 12);
  gtk_container_add (GTK_CONTAINER (gtk_dialog_get_content_area (GTK_DIALOG (dialog))),
                     main_hbox);

  /* Preview Area */
  preview = gimp_drawable_preview_new (drawable, NULL);
  gtk_box_pack_start (GTK_BOX (main_hbox), preview, TRUE, TRUE, 0);

  g_signal_connect (preview, "invalidated", G_CALLBACK (preview_update), NULL);

  /* Terminal */
  terminal = vte_terminal_new ();
  GdkColor foreground, background;
  gdk_color_parse ("blue", &foreground);
  gdk_color_parse ("green", &background);
  vte_terminal_set_color_foreground (VTE_TERMINAL (terminal), &foreground);
  vte_terminal_set_color_background (VTE_TERMINAL (terminal), &background);
  gtk_box_pack_start (GTK_BOX (main_hbox), terminal, TRUE, TRUE, 0);
  
  g_signal_connect (terminal, "child-exited", G_CALLBACK (on_octave_exit), dialog);
  
  char *argv [] = {"octave", "-q", NULL};
  pid = vte_terminal_fork_command (VTE_TERMINAL (terminal), argv [0], argv, NULL,
                                   octave_params.wd, FALSE, FALSE, FALSE);

  /* File Monitoring */
  GFile *monitor_file = g_file_new_for_commandline_arg (octave_params.coutput);
  GFileMonitor *monitor = g_file_monitor_file (monitor_file, G_FILE_MONITOR_NONE,
                                               NULL, NULL);
  g_signal_connect (monitor, "changed", G_CALLBACK (on_file_change), preview);

  gtk_widget_show_all (dialog);
  run = (gimp_dialog_run (GIMP_DIALOG (dialog)) == GTK_RESPONSE_OK);
  gtk_widget_destroy (dialog);
  g_file_monitor_cancel (monitor);

  return run;
}

static
void preview_update (GimpPreview *preview)
{
  GimpDrawable *drawable;
  gint x, y, width, height;
  GimpPixelRgn PR;

  drawable = gimp_drawable_preview_get_drawable (GIMP_DRAWABLE_PREVIEW (preview));

  gimp_pixel_rgn_init (&PR, drawable, 0, 0, drawable->width, drawable->height,
                       TRUE, TRUE);

  //gimp_preview_get_position (preview, &x, &y);
  //gimp_preview_get_size (preview, &width, &height);
  width = drawable->width;
  height = drawable->height;
  x = y = 0;

  if (result_layer != NULL)
    {
      gimp_pixel_rgn_set_rect (&PR, result_layer, x, y, width, height);
      gimp_drawable_preview_draw_region (GIMP_DRAWABLE_PREVIEW (preview), &PR);
    }
}

void reset_default ()
{
  /* Set standard input/output/script strings */
  char *home = getenv ("HOME");
  sprintf (octave_params.wd, "%s/%s", home, WORKING_DIR);
  sprintf (octave_params.cinput, "%s/%s/%s", home, WORKING_DIR, MATRIX_IN);
  sprintf (octave_params.coutput, "%s/%s/%s", home, WORKING_DIR, MATRIX_OUT);
}

